package com.gui;

import java.io.File;
import java.io.IOException;
import java.sql.SQLException;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import swing2swt.layout.BorderLayout;
import org.eclipse.swt.widgets.*;
import com.connect.Messaging;
import com.db.Database;
import com.model.DosboxVersion;
import com.model.OrderingVector;
import com.model.Profile;
import com.model.Settings;
import com.model.Template;
import com.model.conf.CompositeConfiguration;
import com.swtdesigner.SWTImageManager;
import com.util.FileUtils;
import com.util.PlatformUtils;
import com.util.StringUtils;


public final class MainWindow {

    public static final String PROGRAM_VERSION = "0.65";
    public static final String PROGRAM_NAME_FULL = "DOSBox Game Launcher";
    public static final String PROGRAM_NAME = PROGRAM_NAME_FULL + " v" + PROGRAM_VERSION;
    public static final int RO_COLUMN_NAMES = 10;
    public static final int EDIT_COLUMN_NAMES = 10;
    
    public static String[] columnNames;

    private Display display;
    private Shell shell;
    private OrderingVector orderingVector = null;
    private java.util.List<Profile> profilesList;
    private java.util.List<DosboxVersion> dbversionsList;
    private java.util.List<Template> templatesList;
    private Database dbase = null;
    private Settings settings = null;
    private File currentThumbFile = null;
    private int thumbHeight;

    private int[] columnIds;
    private Table template_table;
    private Table dbversion_table;
    private Table profile_table;
    private Text notesField;
    private Composite thumbsToolBar;
    private Menu menu_thumb;
    private Link[] link;
    private Menu runWithSubMenu;
    private Menu setupWithSubMenu;
    private Menu viewLinkSubMenu;
    
	
    public static void main(final String[] args) {
        try {
            MainWindow window = new MainWindow();
            window.open();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    public void open() {
        dbase = Database.getInstance();
        settings = Settings.getInstance();

        display = Display.getDefault();
        createContents();

        Messaging mess = null;
        if (settings.getBooleanValue("communication", "port_enabled")) {
            mess = new Messaging(settings.getIntValue("communication", "port"), this);
            mess.start();
        }

        shell.open();
        shell.layout();
        while (!shell.isDisposed()) {
            if (!display.readAndDispatch()) {
                display.sleep();
            }
        }
        SWTImageManager.dispose();
        display.dispose();

        if (mess != null) {
            mess.close();
        }

        try {
            settings.save();
        } catch (IOException e) {
            GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
        }
        try {
            dbase.shutdown();
        } catch (SQLException e) {
            // nothing we can do
        }
    }
  
    private void createContents() {
        initColumnIds();
        orderingVector = new OrderingVector(settings.getIntValues("gui", "sortcolumn"),
                settings.getBooleanValues("gui", "sortascending"));

        shell = new Shell();
        try {
            profilesList = dbase.readProfilesList(orderingVector.toClause());
            dbversionsList = dbase.readDosboxVersionsList();
            templatesList = dbase.readTemplatesList();
        } catch (SQLException e) {
            GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
        }
        shell.addControlListener(new ControlAdapter() {
            public void controlResized(final ControlEvent event) {
                boolean isMaximized = shell.getMaximized();
                if (!isMaximized) {
                    Rectangle rec = shell.getBounds();
                    settings.setIntValue("gui", "width", rec.width);
                    settings.setIntValue("gui", "height", rec.height);
                    settings.setIntValue("gui", "x", rec.x);
                    settings.setIntValue("gui", "y", rec.y);
                }
                settings.setBooleanValue("gui", "maximized", isMaximized);
            } });
        shell.addControlListener(new ControlAdapter() {
            public void controlMoved(final ControlEvent event) {
                if (!shell.getMaximized()) {
                    Rectangle rec = shell.getBounds();
                    settings.setIntValue("gui", "x", rec.x);
                    settings.setIntValue("gui", "y", rec.y);
                }
            } });
        shell.setLayout(new BorderLayout(0, 0));
        shell.setLocation(settings.getIntValue("gui", "x"), settings.getIntValue("gui", "y"));
        boolean maximized = settings.getBooleanValue("gui", "maximized");
        if (maximized) {
            shell.setMaximized(true);
        } else {
            shell.setSize(settings.getIntValue("gui", "width"), settings.getIntValue("gui", "height"));
        }
        shell.setText(settings.msg("main.title", new Object[] {PROGRAM_VERSION}));

        final Menu menu = new Menu(shell, SWT.BAR);
        shell.setMenuBar(menu);

        final MenuItem fileMenuItem = new MenuItem(menu, SWT.CASCADE);
        fileMenuItem.setText(settings.msg("dialog.main.menu.file"));

        final Menu menu_1 = new Menu(fileMenuItem);
        fileMenuItem.setMenu(menu_1);

        if (PlatformUtils.IS_WINDOWS) {
            final MenuItem dfendImportItem = new MenuItem(menu_1, SWT.NONE);
            dfendImportItem.addSelectionListener(new SelectionAdapter() {
                public void widgetSelected(final SelectionEvent event) {
                    doDFendImport();
                }
            });
            dfendImportItem.setText(settings.msg("dialog.main.menu.importdfendprofiles"));

            new MenuItem(menu_1, SWT.SEPARATOR);
        }

        final MenuItem migrateMenuItem = new MenuItem(menu_1, SWT.NONE);
        migrateMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doMigrate();
            }
        });
        migrateMenuItem.setText(settings.msg("dialog.main.menu.migrateprofiles"));

        final MenuItem exportMenuItem = new MenuItem(menu_1, SWT.NONE);
        exportMenuItem.setText(settings.msg("dialog.main.menu.exportprofileslist"));
        exportMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent e) {
                doExportProfilesList();
            }
        });
        
        final MenuItem confImportItem = new MenuItem(menu_1, SWT.NONE);
        confImportItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doImportConfigfile();
            }
        });
        confImportItem.setText(settings.msg("dialog.main.menu.importprofile"));
        new MenuItem(menu_1, SWT.SEPARATOR);
        
        final MenuItem cleanupDatabaseItem = new MenuItem(menu_1, SWT.NONE);
        cleanupDatabaseItem.setText(settings.msg("dialog.main.menu.databasecleanup"));
        cleanupDatabaseItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                try {
                    if (GeneralPurposeDialogs.confirmMessage(shell, settings.msg("dialog.main.confirm.databasecleanup"))) {
                        int itemsRemoved = dbase.cleanup();
                        GeneralPurposeDialogs.infoMessage(shell, settings.msg("dialog.main.notice.databasecleanupok", new Object[] {itemsRemoved}));
                    }
                } catch (SQLException ex) {
                    GeneralPurposeDialogs.warningMessage(shell, ex.getMessage());
                }
            }
        });
        new MenuItem(menu_1, SWT.SEPARATOR);

        final MenuItem settingsMenuItem = new MenuItem(menu_1, SWT.NONE);
        settingsMenuItem.setText(settings.msg("dialog.main.menu.adjustsettings"));
        settingsMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doOpenSettingsDialog();
            }
        });

        final MenuItem exitMenuItem = new MenuItem(menu_1, SWT.NONE);
        exitMenuItem.setText(settings.msg("dialog.main.menu.exit"));
        exitMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                shell.close();
            }
        });

        final TabFolder tabFolder = new TabFolder(shell, SWT.NONE);

        final TabItem profilesTabItem = new TabItem(tabFolder, SWT.NONE);
        profilesTabItem.setText(settings.msg("dialog.main.profiles"));

        final Composite composite = new Composite(tabFolder, SWT.NONE);
        composite.setLayout(new BorderLayout(0, 0));
        profilesTabItem.setControl(composite);

        final ToolBar toolBar = new ToolBar(composite, SWT.NONE);
        toolBar.setLayoutData(BorderLayout.NORTH);

        SelectionAdapter addProfileAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doAddProfile();
            }
        };
        SelectionAdapter editProfAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doEditProfile();
            }
        };
        SelectionAdapter removeProfAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doRemoveProfile();
            }
        };

        final ToolItem newItemToolItem = new ToolItem(toolBar, SWT.PUSH);
        newItemToolItem.setText(settings.msg("dialog.main.addprofile"));
        newItemToolItem.addSelectionListener(addProfileAdapter);

        final ToolItem newItemToolItem_2 = new ToolItem(toolBar, SWT.PUSH);
        newItemToolItem_2.setText(settings.msg("dialog.main.editprofile"));
        newItemToolItem_2.addSelectionListener(editProfAdapter);

        final ToolItem newItemToolItem_1 = new ToolItem(toolBar, SWT.PUSH);
        newItemToolItem_1.setText(settings.msg("dialog.main.removeprofile"));
        newItemToolItem_1.addSelectionListener(removeProfAdapter);

        final SashForm sashForm = new SashForm(composite, SWT.NONE);

        profile_table = new Table(sashForm, SWT.FULL_SELECTION | SWT.BORDER);

        final Composite informationGroup = new Composite(sashForm, SWT.NONE);
        informationGroup.addControlListener(new ControlAdapter() {
            public void controlResized(final ControlEvent event) {
                settings.setIntValues("gui", "sashweights", sashForm.getWeights());
            }
        });
        informationGroup.setLayout(new BorderLayout(0, 0));
        notesField = new Text(informationGroup, SWT.V_SCROLL | SWT.MULTI | SWT.READ_ONLY | SWT.BORDER | SWT.WRAP);

        final Composite composite_2 = new Composite(informationGroup, SWT.NONE);
        final GridLayout gridLayout = new GridLayout();
        gridLayout.marginWidth = 0;
        gridLayout.marginHeight = 2;
        gridLayout.verticalSpacing = 1;
        composite_2.setLayout(gridLayout);
        composite_2.setLayoutData(BorderLayout.SOUTH);

        link = new Link[4];
        for (int i = 0; i < link.length; i++) {
            link[i] = new Link(composite_2, SWT.NONE);
            link[i].setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
            link[i].setText(" ");
            link[i].addSelectionListener(new SelectionAdapter() {
                public void widgetSelected(final SelectionEvent event) {
                    PlatformUtils.openForBrowsing(event.text);
                }
            });
        }

        thumbHeight = settings.getIntValue("gui", "screenshotsheight");
        ScrolledComposite scrolledComposite = new ScrolledComposite(composite, SWT.BORDER | SWT.H_SCROLL);
        scrolledComposite.setMinHeight(thumbHeight + 20);
        scrolledComposite.setLayoutData(BorderLayout.SOUTH);
        
        thumbsToolBar = new Composite(scrolledComposite, SWT.NONE);
        thumbsToolBar.setLayout(new RowLayout(SWT.HORIZONTAL));
        scrolledComposite.setContent(thumbsToolBar);


        menu_thumb = new Menu(thumbsToolBar);

        final MenuItem thumbMenuItem1 = new MenuItem(menu_thumb, SWT.NONE);
        thumbMenuItem1.setText(settings.msg("dialog.main.thumb.remove"));
        thumbMenuItem1.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doRemoveThumb();
            }
        });
        new MenuItem(menu_thumb, SWT.SEPARATOR);

        final MenuItem thumbMenuItem2 = new MenuItem(menu_thumb, SWT.NONE);
        thumbMenuItem2.setText(settings.msg("dialog.main.thumb.openfolder"));
        thumbMenuItem2.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                PlatformUtils.openDirForViewing(currentThumbFile.getParentFile());
            }
        });

        final MenuItem thumbMenuItem3 = new MenuItem(menu_thumb, SWT.NONE);
        thumbMenuItem3.setText(settings.msg("dialog.main.thumb.refresh"));
        thumbMenuItem3.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                SWTImageManager.dispose();
                displayProfileInformation();
            }
        });

        profile_table.addKeyListener(new KeyAdapter() {
            public void keyPressed(final KeyEvent event) {
                if (event.keyCode == SWT.DEL) {
                    doRemoveProfile();
                } else if (event.keyCode == SWT.INSERT) {
                    doAddProfile();
                } else if (event.keyCode == SWT.F2) {
                    doEditProfile(true);
                } else if (event.stateMask == SWT.CTRL && (Character.toLowerCase(event.keyCode) == 'm')) {
                    doToggleFavoriteProfile();
                } else if (event.stateMask == SWT.CTRL && (Character.toLowerCase(event.keyCode) == 'd')) {
                    doDuplicateProfile();
                }
            }
        });

        profile_table.setLayoutData(BorderLayout.WEST);
        profile_table.addTraverseListener(new TraverseListener() {
            public void keyTraversed(final TraverseEvent event) {
                if ((event.stateMask == SWT.CTRL) && (event.detail == SWT.TRAVERSE_RETURN)) {
                    doEditProfile();
                } else if ((event.stateMask == SWT.SHIFT) && (event.detail == SWT.TRAVERSE_RETURN)) {
                    doRunProfile(true);
                } else if (event.detail == SWT.TRAVERSE_RETURN) {
                    doRunProfile(false);
                } 
            }
        });
        profile_table.addMouseListener(new MouseAdapter() {
            public void mouseDoubleClick(final MouseEvent event) {
                doRunProfile(false);
            }
        });

        profile_table.setLinesVisible(true);
        profile_table.setHeaderVisible(true);

        for (int i = 0; i < columnIds.length; i++) {
            addProfileColumn(columnNames[columnIds[i]], i);
        }

        profile_table.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                displayProfileInformation();
            } });

        final Menu menu_2 = new Menu(profile_table);
        profile_table.setMenu(menu_2);

        final MenuItem runMenuItem = new MenuItem(menu_2, SWT.NONE);
        runMenuItem.setText(settings.msg("dialog.main.profile.run"));
        runMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doRunProfile(false);
            }
        });

        final MenuItem runWithMenuItem = new MenuItem(menu_2, SWT.CASCADE);
        runWithMenuItem.setText(settings.msg("dialog.main.profile.runwith"));

        runWithSubMenu = new Menu(runWithMenuItem);
        runWithMenuItem.setMenu(runWithSubMenu);
        updateProfileDbversionsSubmenu();

        final MenuItem setupMenuItem = new MenuItem(menu_2, SWT.NONE);
        setupMenuItem.setText(settings.msg("dialog.main.profile.setup"));
        setupMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doRunProfile(true);
            }
        });

        final MenuItem setupWithMenuItem = new MenuItem(menu_2, SWT.CASCADE);
        setupWithMenuItem.setText(settings.msg("dialog.main.profile.setupwith"));

        setupWithSubMenu = new Menu(setupWithMenuItem);
        setupWithMenuItem.setMenu(setupWithSubMenu);
        updateSetupDbversionsSubmenu();

        new MenuItem(menu_2, SWT.SEPARATOR);

        final MenuItem openGameFolderItem = new MenuItem(menu_2, SWT.NONE);
        openGameFolderItem.setText(settings.msg("dialog.main.profile.openfolder"));
        openGameFolderItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                int index = profile_table.getSelectionIndex();
                if (index != -1) {
                    Profile prof = profilesList.get(index);
                    int dbversionIndex = DosboxVersion.findById(dbversionsList, prof.getDbversionId());
                    DosboxVersion dbversion = dbversionsList.get(dbversionIndex);
                    CompositeConfiguration compConf = new CompositeConfiguration(dbversion, prof, true, shell);
                    PlatformUtils.openDirForViewing(compConf.getCompositeConf().getCanonicalMainDir());
                }
            }
        });
        
        final MenuItem viewLinkMenuItem = new MenuItem(menu_2, SWT.CASCADE);
        viewLinkMenuItem.setText(settings.msg("dialog.main.profile.view"));
        viewLinkSubMenu = new Menu(viewLinkMenuItem);
        viewLinkMenuItem.setMenu(viewLinkSubMenu);

        new MenuItem(menu_2, SWT.SEPARATOR);

        final MenuItem addMenuItem = new MenuItem(menu_2, SWT.NONE);
        addMenuItem.setText(settings.msg("dialog.main.profile.add"));
        addMenuItem.addSelectionListener(addProfileAdapter);

        final MenuItem editMenuItem = new MenuItem(menu_2, SWT.NONE);
        editMenuItem.setText(settings.msg("dialog.main.profile.edit"));
        editMenuItem.addSelectionListener(editProfAdapter);

        final MenuItem duplicateMenuItem = new MenuItem(menu_2, SWT.NONE);
        duplicateMenuItem.setText(settings.msg("dialog.main.profile.duplicate"));
        duplicateMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doDuplicateProfile();
            }
        });

        final MenuItem removeMenuItem = new MenuItem(menu_2, SWT.NONE);
        removeMenuItem.setText(settings.msg("dialog.main.profile.remove"));
        removeMenuItem.addSelectionListener(removeProfAdapter);

        new MenuItem(menu_2, SWT.SEPARATOR);

        final MenuItem toggleMenuItem = new MenuItem(menu_2, SWT.NONE);
        toggleMenuItem.setText(settings.msg("dialog.main.profile.togglefavorite"));
        toggleMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doToggleFavoriteProfile();
            }
        });

        if (PlatformUtils.IS_WINDOWS) {
            new MenuItem(menu_2, SWT.SEPARATOR);
            final MenuItem shortcutMenuItem = new MenuItem(menu_2, SWT.NONE);
            shortcutMenuItem.setText(settings.msg("dialog.main.profile.createshortcut"));
            shortcutMenuItem.addSelectionListener(new SelectionAdapter() {
                public void widgetSelected(final SelectionEvent event) {
                    doCreateShortcut();
                }
            });
        }

        menu_2.addMenuListener(new MenuAdapter() {
            public void menuShown(final MenuEvent event) {
                if (profile_table.getSelectionIndex() != -1) {
                    boolean hasSetup = profilesList.get(profile_table.getSelectionIndex()).hasSetup();
                    setupMenuItem.setEnabled(hasSetup);
                    setupWithMenuItem.setEnabled(hasSetup);
                }
            }
        });

        sashForm.setWeights(settings.getIntValues("gui", "sashweights"));

        final TabItem dosboxTabItem = new TabItem(tabFolder, SWT.NONE);
        dosboxTabItem.setText(settings.msg("dialog.main.dosboxversions"));

        final Composite composite_1 = new Composite(tabFolder, SWT.NONE);
        composite_1.setLayout(new BorderLayout(0, 0));
        dosboxTabItem.setControl(composite_1);

        final ToolBar toolBar_1 = new ToolBar(composite_1, SWT.NONE);
        toolBar_1.setLayoutData(BorderLayout.NORTH);

        SelectionAdapter addDosboxAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doAddDosboxVersion();
            }
        };
        SelectionAdapter editDosboxAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doEditDosboxVersion();
            }
        };
        SelectionAdapter removeDbAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doRemoveDosboxVersion();
            }
        };

        final ToolItem newItemToolItem_3 = new ToolItem(toolBar_1, SWT.PUSH);
        newItemToolItem_3.addSelectionListener(addDosboxAdapter);
        newItemToolItem_3.setText(settings.msg("dialog.main.addversion"));

        final ToolItem newItemToolItem_4 = new ToolItem(toolBar_1, SWT.PUSH);
        newItemToolItem_4.addSelectionListener(editDosboxAdapter);
        newItemToolItem_4.setText(settings.msg("dialog.main.editversion"));

        final ToolItem newItemToolItem_5 = new ToolItem(toolBar_1, SWT.PUSH);
        newItemToolItem_5.addSelectionListener(removeDbAdapter);
        newItemToolItem_5.setText(settings.msg("dialog.main.removeversion"));

        dbversion_table = new Table(composite_1, SWT.FULL_SELECTION | SWT.BORDER);
        dbversion_table.addKeyListener(new KeyAdapter() {
            public void keyPressed(final KeyEvent event) {
                if (event.keyCode == SWT.DEL) {
                    doRemoveDosboxVersion();
                } else if (event.keyCode == SWT.INSERT) {
                    doAddDosboxVersion();
                } else if (event.stateMask == SWT.CTRL && (Character.toLowerCase(event.keyCode) == 'm')) {
                    doToggleDefaultVersion();
                }
            }
        });
        dbversion_table.addTraverseListener(new TraverseListener() {
            public void keyTraversed(final TraverseEvent event) {
                if ((event.stateMask == SWT.CTRL) && (event.detail == SWT.TRAVERSE_RETURN)) {
                    doEditDosboxVersion();
                } else if (event.detail == SWT.TRAVERSE_RETURN) {
                    doRunDosbox();
                }
            }
        });
        dbversion_table.addMouseListener(new MouseAdapter() {
            public void mouseDoubleClick(final MouseEvent event) {
                doRunDosbox();
            }
        });

        dbversion_table.setLinesVisible(true);
        dbversion_table.setHeaderVisible(true);

        addDBColumn(settings.msg("dialog.main.dosboxversions.column.title"), 0);
        addDBColumn(settings.msg("dialog.main.dosboxversions.column.version"), 1);
        addDBColumn(settings.msg("dialog.main.dosboxversions.column.path"), 2);
        addDBColumn(settings.msg("dialog.main.dosboxversions.column.default"), 3);
        addDBColumn(settings.msg("dialog.main.dosboxversions.column.id"), 4);

        final Menu menu_4 = new Menu(dbversion_table);
        dbversion_table.setMenu(menu_4);

        final MenuItem runDBMenuItem = new MenuItem(menu_4, SWT.NONE);
        runDBMenuItem.setText(settings.msg("dialog.main.dosboxversion.run"));
        runDBMenuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doRunDosbox();
            }
        });

        new MenuItem(menu_4, SWT.SEPARATOR);

        final MenuItem addDBMenuItem = new MenuItem(menu_4, SWT.NONE);
        addDBMenuItem.setText(settings.msg("dialog.main.dosboxversion.add"));
        addDBMenuItem.addSelectionListener(addDosboxAdapter);

        final MenuItem editDBMenuItem = new MenuItem(menu_4, SWT.NONE);
        editDBMenuItem.setText(settings.msg("dialog.main.dosboxversion.edit"));
        editDBMenuItem.addSelectionListener(editDosboxAdapter);

        final MenuItem removeDBMenuItem = new MenuItem(menu_4, SWT.NONE);
        removeDBMenuItem.setText(settings.msg("dialog.main.dosboxversion.remove"));
        removeDBMenuItem.addSelectionListener(removeDbAdapter);

        new MenuItem(menu_4, SWT.SEPARATOR);

        final MenuItem toggleDefaultItem = new MenuItem(menu_4, SWT.NONE);
        toggleDefaultItem.setText(settings.msg("dialog.main.dosboxversion.toggledefault"));
        toggleDefaultItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doToggleDefaultVersion();
            }
        });

        new MenuItem(menu_4, SWT.SEPARATOR);

        final MenuItem openFolderItem = new MenuItem(menu_4, SWT.NONE);
        openFolderItem.setText(settings.msg("dialog.main.dosboxversion.openfolder"));
        openFolderItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                int index = dbversion_table.getSelectionIndex();
                if (index != -1) {
                    PlatformUtils.openDirForViewing(dbversionsList.get(index).getCanonicalConfFile().getParentFile());
                }
            }
        });

        final TabItem templatesTabItem = new TabItem(tabFolder, SWT.NONE);
        templatesTabItem.setText(settings.msg("dialog.main.templates"));

        final Composite composite_1_1 = new Composite(tabFolder, SWT.NONE);
        composite_1_1.setLayout(new BorderLayout(0, 0));
        templatesTabItem.setControl(composite_1_1);

        final ToolBar toolBar_1_1 = new ToolBar(composite_1_1, SWT.NONE);
        toolBar_1_1.setLayoutData(BorderLayout.NORTH);

        SelectionAdapter addTemplAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doAddTemplate();
            }
        };
        SelectionAdapter editTemplAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doEditTemplate();
            }
        };
        SelectionAdapter remTemplAdapter = new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doRemoveTemplate();
            }
        };

        final ToolItem newItemToolItem_6 = new ToolItem(toolBar_1_1, SWT.PUSH);
        newItemToolItem_6.setText(settings.msg("dialog.main.addtemplate"));
        newItemToolItem_6.addSelectionListener(addTemplAdapter);

        final ToolItem newItemToolItem_7 = new ToolItem(toolBar_1_1, SWT.PUSH);
        newItemToolItem_7.setText(settings.msg("dialog.main.edittemplate"));
        newItemToolItem_7.addSelectionListener(editTemplAdapter);

        final ToolItem newItemToolItem_8 = new ToolItem(toolBar_1_1, SWT.PUSH);
        newItemToolItem_8.setText(settings.msg("dialog.main.removetemplate"));
        newItemToolItem_8.addSelectionListener(remTemplAdapter);

        template_table = new Table(composite_1_1, SWT.FULL_SELECTION | SWT.BORDER);
        template_table.addKeyListener(new KeyAdapter() {
            public void keyPressed(final KeyEvent event) {
                if (event.keyCode == SWT.DEL) {
                    doRemoveTemplate();
                } else if (event.keyCode == SWT.INSERT) {
                    doAddTemplate();
                } else if (event.stateMask == SWT.CTRL && (Character.toLowerCase(event.keyCode) == 'm')) {
                    doToggleDefaultTemplate();
                }
            }
        });
        template_table.addTraverseListener(new TraverseListener() {
            public void keyTraversed(final TraverseEvent event) {
                if ((event.stateMask == SWT.CTRL) && (event.detail == SWT.TRAVERSE_RETURN)) {
                    doEditTemplate();
                }
            }
        });

        template_table.setLinesVisible(true);
        template_table.setHeaderVisible(true);

        addTemplateColumn(settings.msg("dialog.main.templates.column.title"), 0);
        addTemplateColumn(settings.msg("dialog.main.templates.column.default"), 1);

        final Menu menu_3 = new Menu(template_table);
        template_table.setMenu(menu_3);

        final MenuItem addTemplItem = new MenuItem(menu_3, SWT.NONE);
        addTemplItem.setText(settings.msg("dialog.main.template.add"));
        addTemplItem.addSelectionListener(addTemplAdapter);

        final MenuItem editTemplItem = new MenuItem(menu_3, SWT.NONE);
        editTemplItem.setText(settings.msg("dialog.main.template.edit"));
        editTemplItem.addSelectionListener(editTemplAdapter);

        final MenuItem removeTemplItem = new MenuItem(menu_3, SWT.NONE);
        removeTemplItem.setText(settings.msg("dialog.main.template.remove"));
        removeTemplItem.addSelectionListener(remTemplAdapter);

        new MenuItem(menu_3, SWT.SEPARATOR);

        final MenuItem toggleTemplItem = new MenuItem(menu_3, SWT.NONE);
        toggleTemplItem.setText(settings.msg("dialog.main.template.toggledefault"));
        toggleTemplItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doToggleDefaultTemplate();
            }
        });

        // init values
        for (Profile prof: profilesList) {
            addProfileToTable(prof);
        }
        for (DosboxVersion dbversion: dbversionsList) {
            addDosboxVersionToTable(dbversion);
        }
        for (Template template: templatesList) {
            addTemplateToTable(template);
        }

        profile_table.setSortColumn(profile_table.getColumn(findColumnById(settings.getIntValues("gui", "sortcolumn")[0])));
        profile_table.setSortDirection(settings.getBooleanValues("gui", "sortascending")[0]? SWT.UP: SWT.DOWN);
        profile_table.setColumnOrder(settings.getIntValues("gui", "columnorder"));
        profile_table.setSelection(settings.getIntValue("gui", "selectedprofile"));
        profile_table.setFocus();
        displayProfileInformation();
    }

    private void doReorderProfiles(final int columnId, final SelectionEvent event) {
        TableColumn sortColumn = profile_table.getSortColumn();
        TableColumn currentColumn = (TableColumn) event.widget;
        int dir = profile_table.getSortDirection();
        if (sortColumn.equals(currentColumn)) {
            dir = dir == SWT.UP ? SWT.DOWN : SWT.UP;
        } else {
            profile_table.setSortColumn(currentColumn);
            dir = SWT.UP;
        }
        int currentProfileId = -1;
        if (profile_table.getSelectionIndex() != -1) {
            currentProfileId = profilesList.get(profile_table.getSelectionIndex()).getId();
        }
        try {
            orderingVector.addOrdering(columnIds[columnId], dir == SWT.UP);
            profilesList = dbase.readProfilesList(orderingVector.toClause());
        } catch (SQLException ex) {
            GeneralPurposeDialogs.warningMessage(shell, ex.getMessage());
        }
        for (int i = 0; i < profilesList.size(); i++) {
            setTableItem(profile_table.getItem(i), profilesList.get(i));
        }
        profile_table.setSortDirection(dir);
        if (currentProfileId != -1) {
            profile_table.setSelection(Profile.findById(profilesList, currentProfileId));
        }
        settings.setIntValues("gui", "sortcolumn", orderingVector.getColumns());
        settings.setBooleanValues("gui", "sortascending", orderingVector.getAscendings());
    }

    private void displayNotes(final String notes) {
        notesField.setText(notes);
    }

    private void displayLinks(final String[] p_link, final String[] p_linkTitle) {
        for (int i = 0; i < link.length; i++) {
            if (p_link[i] == null || "".equals(p_link[i])) {
                link[i].setText(" ");
                link[i].setToolTipText(null);
            } else {
                String url = p_link[i];
                String tag = p_link[i];
                if (url.indexOf("://") == -1) {
                    url = "file://" + FileUtils.canonicalToData(url).getPath();
                    tag = FileUtils.makeRelativeToDosroot(FileUtils.canonicalToData(tag)).getPath();
                }
                if (p_linkTitle[i] != null && !"".equals(p_linkTitle[i]))
                    tag = p_linkTitle[i];
                StringBuffer text = new StringBuffer("<a href=\"").append(url).append("\">").append(tag).append("</a>");
                link[i].setText(text.toString());
                link[i].setToolTipText(url);
            }
        }
    }

    private void displayScreenshots(final Profile prof) {
        for (Control c: thumbsToolBar.getChildren()) {
            c.setMenu(null);
            c.dispose();
        }
        thumbsToolBar.pack();
        if (prof != null) {
            File path = prof.getCanonicalCaptures();
            File[] files = path.listFiles();
            if (files != null) {
                for (File file: files) {
                    String label = file.getName().toLowerCase();
                    if (label.endsWith(".png")) {
                        label = ' ' + label.substring(0, label.lastIndexOf(".png")) + ' ';
                        final Button buttonItem = new Button(thumbsToolBar, SWT.PUSH);
                        buttonItem.setToolTipText(file.getPath());
                        buttonItem.setImage(SWTImageManager.getResizedImage(display, thumbHeight, file.getPath(), label));
                        buttonItem.pack();
                        thumbsToolBar.pack();
                        buttonItem.addSelectionListener(new SelectionAdapter() {
                            public void widgetSelected(final SelectionEvent event) {
                                final Thumb thumbDialog = new Thumb(shell);
                                thumbDialog.setThumb(buttonItem.getToolTipText());
                                thumbDialog.open();
                            }
                        });
                        buttonItem.addMouseListener(new MouseAdapter() {
                            public void mouseDown(MouseEvent arg0) {
                                currentThumbFile = new File(buttonItem.getToolTipText());
                            }
                        });
                        buttonItem.setMenu(menu_thumb);
                    }
                }
            }
        }
        thumbsToolBar.setVisible(thumbsToolBar.getChildren().length != 0);
        thumbsToolBar.pack();
    }

    private void doAddProfile() {
        doAddProfile(null);
    }

    private void doAddProfile(final String file) {
        if (checkDefaultDBVersion() == null) {
            return;
        }
        final EditProfileDialog addProfileDialog = new EditProfileDialog(shell);
        if (file != null) {
            addProfileDialog.sendToProfile(file);
        }
        updateWithAddedProfile((Profile)addProfileDialog.open());
    }

    public void addProfile(final String file) {
        display.syncExec(new Runnable() {
            public void run() {
                doAddProfile(file);
            }
        });
    }

    private void doAddDosboxVersion() {
        final EditDosboxVersionDialog addVersionDialog = new EditDosboxVersionDialog(shell);
        DosboxVersion dbversion = (DosboxVersion)addVersionDialog.open();
        if (dbversion != null) {
            updateDosboxVersionList(dbversion);
        }
    }

    private void doAddTemplate() {
        if (checkDefaultDBVersion() == null) {
            return;
        }
        final EditTemplateDialog addTemplateDialog = new EditTemplateDialog(shell);
        Template template = (Template)addTemplateDialog.open();
        if (template != null) {
            updateTemplateList(template);
        }
    }

    private void doToggleFavoriteProfile() {
        int index = profile_table.getSelectionIndex();
        if (index != -1) {
            Profile profile = profilesList.get(index);
            try {
                dbase.toggleFavorite(profile.getId());
            } catch (SQLException e) {
                GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
            }
            profile.toggleDefault();
            profilesList.set(index, profile);
            setTableItem(profile_table.getItem(index), profile);
        }
    }

    private void doToggleDefaultVersion() {
        int index = dbversion_table.getSelectionIndex();
        if (index != -1) {
            DosboxVersion ver = dbversionsList.get(index);
            ver.toggleDefault();
            try {
                dbase.addOrEditDosboxVersion(ver.getTitle(), ver.getPath(), ver.isMultiConfig(), ver.isDefault(),
                        ver.getParameters(), ver.getVersion(), ver.getId());
            } catch (SQLException e) {
                GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
            }
            updateDosboxVersionList(ver);
        }
    }

    private void doToggleDefaultTemplate() {
        int index = template_table.getSelectionIndex();
        if (index != -1) {
            Template temp = templatesList.get(index);
            temp.toggleDefault();
            try {
                dbase.addOrEditTemplate(temp.getTitle(), temp.getdbversionId(), temp.isDefault(), temp.getId());
            } catch (SQLException e) {
                GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
            }
            updateTemplateList(temp);
        }
    }

    private void doEditProfile() {
        doEditProfile(false);
    }

    private void doEditProfile(final boolean focusTitle) {
        int index = profile_table.getSelectionIndex();
        if (index != -1) {
            final EditProfileDialog editProfileDialog = new EditProfileDialog(shell);
            editProfileDialog.setProfile(profilesList.get(index));
            if (focusTitle) {
                editProfileDialog.focusTitle();
            }
            Profile profile = (Profile)editProfileDialog.open();
            if (profile != null) {
                boolean quickUpdate = true;
                if (settings.getBooleanValue("gui", "autosortonupdate")) {
                    try {
                        profilesList = dbase.readProfilesList(orderingVector.toClause());
                        if (index != Profile.findById(profilesList, profile.getId())) {
                            quickUpdate = false;
                        }
                    } catch (SQLException e) {
                        GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
                    }
                }
                if (quickUpdate) {
                    profilesList.set(index, profile);
                    setTableItem(profile_table.getItem(index), profile);
                } else {
                    updateProfilesList(profile);
                }
                displayProfileInformation();
            }
        }
    }

    private void doDuplicateProfile() {
        int index = profile_table.getSelectionIndex();
        if (index != -1) {
            Profile orgProf = profilesList.get(index);
            updateWithAddedProfile(EditProfileDialog.duplicateProfile(orgProf, dbversionsList, dbase, shell));
        }
    }

    private void updateWithAddedProfile(final Profile profile) {
        if (profile != null) {
            if (settings.getBooleanValue("gui", "autosortonupdate")) {
                updateProfilesList(profile);
            } else {
                profilesList.add(profile);
                addProfileToTable(profile);
                profile_table.setSelection(profile_table.getItemCount() - 1);
                profile_table.setFocus();
            }
            displayProfileInformation();
        }
    }

    private void doEditDosboxVersion() {
        int index = dbversion_table.getSelectionIndex();
        if (index != -1) {
            final EditDosboxVersionDialog editVersionDialog = new EditDosboxVersionDialog(shell);
            editVersionDialog.setDosboxVersion(dbversionsList.get(index));
            DosboxVersion dbversion = (DosboxVersion)editVersionDialog.open();
            if (dbversion != null) {
                updateDosboxVersionList(dbversion);
            }
        }
    }

    private void updateProfilesList(final Profile profile) {
        try {
            profilesList = dbase.readProfilesList(orderingVector.toClause());
        } catch (SQLException e) {
            GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
        }
        profile_table.setRedraw(false);
        profile_table.removeAll();
        for (Profile prof: profilesList) {
            addProfileToTable(prof);
        }
        if (profile != null) {
            profile_table.setSelection(Profile.findById(profilesList, profile.getId()));
        }
        profile_table.setRedraw(true);
        profile_table.setFocus();
    }

    private void updateDosboxVersionList(final DosboxVersion dbversion) {
        try {
            dbversionsList = dbase.readDosboxVersionsList();
        } catch (SQLException e) {
            GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
        }
        dbversion_table.removeAll();
        for (DosboxVersion version: dbversionsList) {
            addDosboxVersionToTable(version);
        }
        dbversion_table.setSelection(DosboxVersion.findById(dbversionsList, dbversion.getId()));
        dbversion_table.setFocus();

        updateProfileDbversionsSubmenu();
        updateSetupDbversionsSubmenu();
    }

    private void updateTemplateList(final Template template) {
        try {
            templatesList = dbase.readTemplatesList();
        } catch (SQLException e) {
            GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
        }
        template_table.removeAll();
        for (Template temp: templatesList) {
            addTemplateToTable(temp);
        }
        template_table.setSelection(Template.findById(templatesList, template.getId()));
        template_table.setFocus();
    }

    private void doEditTemplate() {
        int index = template_table.getSelectionIndex();
        if (index != -1) {
            final EditTemplateDialog editTemplDialog = new EditTemplateDialog(shell);
            editTemplDialog.setTemplate(templatesList.get(index));
            Template template = (Template)editTemplDialog.open();
            if (template != null) {
                updateTemplateList(template);
            }
        }
    }

    private void doRemoveProfile() {
        int index = profile_table.getSelectionIndex();
        if ((index != -1) && GeneralPurposeDialogs.confirmMessage(shell, settings.msg("dialog.main.confirm.removeprofile"))) {
            Profile prof = profilesList.get(index);
            try {
                dbase.removeProfile(prof.getId());
                profile_table.remove(index);
                profilesList.remove(index);
                displayProfileInformation();
                SWTImageManager.dispose();
                if (GeneralPurposeDialogs.confirmMessage(shell, settings.msg("dialog.main.confirm.removeprofileconf", 
                        new Object[] {prof.getConfPathAndFile()}))) {
                    FileUtils.removeFile(prof.getCanonicalConfFile());
                }
                if (GeneralPurposeDialogs.confirmMessage(shell, settings.msg("dialog.main.confirm.removeprofilecaptures", 
                        new Object[] {prof.getCaptures()}))) {
                    FileUtils.removeFilesInDirAndDir(prof.getCanonicalCaptures());
                }
            } catch (SQLException e) {
                GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
            }
        }
    }

    private void doRemoveDosboxVersion() {
        int index = dbversion_table.getSelectionIndex();
        if ((index != -1) && GeneralPurposeDialogs.confirmMessage(shell, settings.msg("dialog.main.confirm.removedosboxversion"))) {
            try {
                dbase.removeDosboxVersion((dbversionsList.get(index)).getId());
                dbversion_table.remove(index);
                dbversionsList.remove(index);
                updateProfileDbversionsSubmenu();
                updateSetupDbversionsSubmenu();
            } catch (SQLException e) {
                GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
            }
        }
    }

    private void doRemoveTemplate() {
        int index = template_table.getSelectionIndex();
        if ((index != -1) && GeneralPurposeDialogs.confirmMessage(shell, settings.msg("dialog.main.confirm.removetemplate"))) {
            int templateId = (templatesList.get(index)).getId();
            try {
                dbase.removeTemplate(templateId);
                template_table.remove(index);
                templatesList.remove(index);
                File conffile = FileUtils.constructCanonicalTemplateFileLocation(templateId);
                if (GeneralPurposeDialogs.confirmMessage(shell, settings.msg("dialog.main.confirm.removetemplateconf", new Object[] {conffile}))) {
                    FileUtils.removeFile(conffile);
                }
            } catch (SQLException e) {
                GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
            }
        }
    }

    private void addProfileToTable(final Profile prof) {
        final TableItem newItemTableItem = new TableItem(profile_table, SWT.BORDER); 
        setTableItem(newItemTableItem, prof);
    }

    private void addDosboxVersionToTable(final DosboxVersion dbversion) {
        final TableItem newItemTableItem = new TableItem(dbversion_table, SWT.BORDER); 
        setTableItem(newItemTableItem, dbversion);
    }

    private void addTemplateToTable(final Template template) {
        final TableItem newItemTableItem = new TableItem(template_table, SWT.BORDER); 
        setTableItem(newItemTableItem, template);
    }

    private void setTableItem(final TableItem newItemTableItem, final Profile prof) {
        for (int i = 0; i < columnIds.length; i++) {
            String value;
            switch (columnIds[i]) {
                case 0: value = prof.getTitle(); break;
                case 1: value = prof.hasSetupString(); break;
                case 2: value = prof.getDeveloperName(); break;
                case 3: value = prof.getPublisherName(); break;
                case 4: value = prof.getGenre(); break;
                case 5: value = prof.getYear(); break;
                case 6: value = prof.getStatus(); break;
                case 7: value = prof.isDefaultString(); break;
                case 8: value = String.valueOf(prof.getId()); break;
                case 9: value = String.valueOf(prof.getDbversionId()); break;
                case 10: case 11: case 12: case 13: case 14: case 15: case 16: case 17:
                    value = prof.getCustomString(columnIds[i] - RO_COLUMN_NAMES); break;
                case 18:value = prof.getCustomInt(0) + " %"; break;
                case 19:value = String.valueOf(prof.getCustomInt(1)); break;
                default:value = "";
            }
            newItemTableItem.setText(i, value);
        }
    }

    private void setTableItem(final TableItem newItemTableItem, final DosboxVersion dbversion) {
        newItemTableItem.setText(0, dbversion.getTitle());
        newItemTableItem.setText(1, dbversion.getVersion());
        newItemTableItem.setText(2, dbversion.getPath());
        newItemTableItem.setText(3, dbversion.isDefaultString());
        newItemTableItem.setText(4, String.valueOf(dbversion.getId()));
    }

    private void setTableItem(final TableItem newItemTableItem, final Template template) {
        newItemTableItem.setText(0, template.getTitle());
        newItemTableItem.setText(1, template.isDefaultString());
    }

    private void doRunProfile(final boolean setup) {
        int index = profile_table.getSelectionIndex();
        if (index != -1) {
            Profile prof = profilesList.get(index);
            if (!setup || prof.hasSetup()) {
                try {
                    FileUtils.doRunProfile(prof, dbversionsList, setup);
                } catch (IOException e) {
                    GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
                }
            }
        }
    }

    private void doRunDosbox() {
        int index = dbversion_table.getSelectionIndex();
        if (index != -1) {
            try {
                FileUtils.doRunDosbox(dbversionsList.get(index));
            } catch (IOException e) {
                GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
            }
        }
    }

    private DosboxVersion checkDefaultDBVersion() {
        int defaultVersionIdx = DosboxVersion.findDefault(dbversionsList);
        if (defaultVersionIdx == -1) {
            GeneralPurposeDialogs.infoMessage(shell,settings.msg("dialog.main.required.defaultdosboxversion"));
            return null;
        }
        return dbversionsList.get(defaultVersionIdx);
    }

    private void doDFendImport() {
        DosboxVersion defaultDbversion = checkDefaultDBVersion();
        if (defaultDbversion == null) {
            return;
        }

        if (settings.getIntValue("profiledefaults", "confpath") == 1) {
            GeneralPurposeDialogs.infoMessage(shell, settings.msg("dialog.main.notice.dfendimportconflocation",
                    new Object[] {SettingsDialog.confLocations[0]}));
        }

        DFendImportDialog importDialog = new DFendImportDialog(shell);
        importDialog.setDefaultDosboxVersion(defaultDbversion);
        if (importDialog.open() != null) {
            try {
                profilesList = dbase.readProfilesList(orderingVector.toClause());
                profile_table.removeAll();
                for (Profile prof: profilesList) {
                    addProfileToTable(prof);
                }
            } catch (SQLException e) {
                GeneralPurposeDialogs.warningMessage(shell, e.getMessage());
            }
        }
    }

    private void doMigrate() {
        GeneralPurposeDialogs.infoMessage(shell, settings.msg("dialog.main.notice.premigration"));
        String from = (String) new MigrateDialog(shell).open();
        if (from != null) {
            GeneralPurposeDialogs.infoMessage(shell, settings.msg("dialog.main.notice.postmigration",
                    new Object[] {from, FileUtils.getDosRoot()}));
        }
    }

    private void displayProfileInformation() {
        int index = profile_table.getSelectionIndex();
        if (index == -1) {
            displayScreenshots(null);
            displayNotes("");
            displayLinks(new String[] {null, null, null, null}, new String[] {null, null, null, null});
            updateViewLinkSubmenu(null);
        } else {
            settings.setIntValue("gui", "selectedprofile", index);
            Profile prof = profilesList.get(index);
            displayScreenshots(prof);
            displayNotes(prof.getNotes());
            displayLinks(prof.getLinks(), prof.getLinkTitles());
            updateViewLinkSubmenu(prof);
        }
    }

    private void doRemoveThumb() {
        if (GeneralPurposeDialogs.confirmMessage(shell, settings.msg("dialog.main.confirm.removethumb", new Object[] {currentThumbFile}))) {
            FileUtils.removeFile(currentThumbFile);
            displayProfileInformation();
        }
        currentThumbFile = null;
    }

    private void doCreateShortcut() {
        int index = profile_table.getSelectionIndex();
        if (index != -1) {
            PlatformUtils.createShortcut(profilesList.get(index), dbversionsList);
        }
    }

    private void addProfileColumn(final String title, final int colIndex) {
        final String width = "column" + (columnIds[colIndex] + 1) + "width";
        final TableColumn column = new TableColumn(profile_table, SWT.NONE);
        column.setWidth(settings.getIntValue("gui", width));
        column.setMoveable(true);
        column.setText(title);
        if ((columnIds[colIndex] == 8) || (columnIds[colIndex] == 9) ||
                (columnIds[colIndex] == 18) || (columnIds[colIndex] == 19)) { // numeric values
            column.setAlignment(SWT.RIGHT);
        }
        column.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(final SelectionEvent event) {
                doReorderProfiles(colIndex, event);
            }
        });
        column.addControlListener(new ControlAdapter() {
            public void controlResized(final ControlEvent event) {
                settings.setIntValue("gui", width, column.getWidth());
            }
            public void controlMoved(final ControlEvent event) {
                settings.setIntValues("gui", "columnorder", profile_table.getColumnOrder());
            }
        });
    }

    private void addDBColumn(final String title, final int colIndex) {
        final String width = "column2_" + (colIndex + 1) + "width";
        final TableColumn column = new TableColumn(dbversion_table, SWT.NONE);
        column.setWidth(settings.getIntValue("gui", width));
        column.setText(title);
        column.addControlListener(new ControlAdapter() {
            public void controlResized(final ControlEvent event) {
                settings.setIntValue("gui", width, column.getWidth());
            }
        });
    }

    private void addTemplateColumn(final String title, final int colIndex) {
        final String width = "column3_" + (colIndex + 1) + "width";
        final TableColumn column = new TableColumn(template_table, SWT.NONE);
        column.setWidth(settings.getIntValue("gui", width));
        column.setText(title);
        column.addControlListener(new ControlAdapter() {
            public void controlResized(final ControlEvent event) {
                settings.setIntValue("gui", width, column.getWidth());
            }
        });
    }

    private void doOpenSettingsDialog() {
        SettingsDialog sDialog = new SettingsDialog(shell);
        if (((Boolean)sDialog.open())) {
            // first, store current table properties
            int index = profile_table.getSelectionIndex();
            int nrColumns = profile_table.getColumnCount();
            String order = settings.getValue("gui", "columnorder");

            // then clean the table, reset column order and dispose all columns
            profile_table.removeAll();
            StringBuffer cols = new StringBuffer();
            for (int i = 0; i < nrColumns; i++) {
                cols.append(i).append(' ');
            }
            profile_table.setColumnOrder(StringUtils.stringToIntArray(cols.toString()));
            for (int i = nrColumns - 1; i >= 0; i--) {
                profile_table.getColumn(i).dispose();
            }

            // next, recreate column ids and add selected columns to table
            initColumnIds();
            for (int i = 0; i < columnIds.length; i++) {
                addProfileColumn(columnNames[columnIds[i]], i);
            }

            // redefine sort column in table
            profile_table.setSortColumn(profile_table.getColumn(findColumnById(settings.getIntValues("gui", "sortcolumn")[0])));
            profile_table.setSortDirection(settings.getBooleanValues("gui", "sortascending")[0]? SWT.UP: SWT.DOWN);

            // restore saved columnorder since disposing and adding columns invalidates that
            settings.setValue("gui", "columnorder", order);

            // also reset orderingvector
            orderingVector = new OrderingVector(settings.getIntValues("gui", "sortcolumn"),
                    settings.getBooleanValues("gui", "sortascending"));

            // finally, refill table with data
            Profile prof = (index == -1)? null: profilesList.get(index);
            updateProfilesList(prof);
        }
    }

    private void doImportConfigfile() {
        FileDialog dialog = new FileDialog (shell, SWT.OPEN);
        dialog.setFilterNames (new String [] { settings.msg("filetype.conf"),
                settings.msg("filetype.exe") + ", " + settings.msg("filetype.booterimage"), FileUtils.ALL_FILTER});
        dialog.setFilterExtensions (new String [] {FileUtils.CNF_FILTER, 
                FileUtils.EXE_FILTER + ";" + FileUtils.BTR_FILTER, FileUtils.ALL_FILTER});
        String result = dialog.open();
        if (result != null) {
            doAddProfile(result);
        }
    }

    private void initColumnIds() {
        columnNames = new String[RO_COLUMN_NAMES + EDIT_COLUMN_NAMES];
        columnNames[0] = settings.msg("dialog.main.profiles.column.title");
        columnNames[1] = settings.msg("dialog.main.profiles.column.setup");
        columnNames[2] = settings.msg("dialog.main.profiles.column.developer");
        columnNames[3] = settings.msg("dialog.main.profiles.column.publisher");
        columnNames[4] = settings.msg("dialog.main.profiles.column.genre");
        columnNames[5] = settings.msg("dialog.main.profiles.column.year");
        columnNames[6] = settings.msg("dialog.main.profiles.column.status");
        columnNames[7] = settings.msg("dialog.main.profiles.column.favorite");
        columnNames[8] = settings.msg("dialog.main.profiles.column.id");
        columnNames[9] = settings.msg("dialog.main.profiles.column.dosboxversionid");
        for (int i = 0; i < EDIT_COLUMN_NAMES; i++) {
            columnNames[i + RO_COLUMN_NAMES] = settings.getValue("gui", "custom" + (i + 1));
        }

        int amount = 0;
        for (int i = 0; i < columnNames.length; i++) {
            if (settings.getBooleanValue("gui", "column" + (i + 1) + "visible")) {
                amount++;
            }
        }

        int cNr = 0;
        columnIds = new int[amount];
        for (int i = 0; i < columnNames.length; i++) {
            if (settings.getBooleanValue("gui", "column" + (i + 1) + "visible")) {
                columnIds[cNr++] = i;
            }
        }
    }

    private int findColumnById(final int id) {
        for (int i = 0; i < columnIds.length; i++) {
            if (columnIds[i] == id) {
                return i;
            }
        }
        return -1;
    }

    private void updateProfileDbversionsSubmenu() {
        MenuItem parent = runWithSubMenu.getParentItem();
        runWithSubMenu.dispose();
        runWithSubMenu = new Menu(parent);
        parent.setMenu(runWithSubMenu);
        for (int i = 0; i < dbversionsList.size(); i++) {
            final MenuItem runWithMenuItem = new MenuItem(runWithSubMenu, SWT.NONE);
            runWithMenuItem.setText(dbversionsList.get(i).getTitle());
            runWithMenuItem.addSelectionListener(new SelectionAdapter() {
                public void widgetSelected(final SelectionEvent event) {
                    int index = profile_table.getSelectionIndex();
                    if (index != -1) {
                        Profile prof = profilesList.get(index);
                        DosboxVersion dbversion = dbversionsList.get(runWithSubMenu.indexOf(runWithMenuItem));
                        try {
                            FileUtils.doRunProfile(prof, dbversion, false);
                        } catch (IOException exc) {
                            GeneralPurposeDialogs.warningMessage(shell, exc.getMessage());
                        }
                    }
                }
            });
        }
    }

    private void updateSetupDbversionsSubmenu() {
        MenuItem parent = setupWithSubMenu.getParentItem();
        setupWithSubMenu.dispose();
        setupWithSubMenu = new Menu(parent);
        parent.setMenu(setupWithSubMenu);
        for (int i = 0; i < dbversionsList.size(); i++) {
            final MenuItem setupWithMenuItem = new MenuItem(setupWithSubMenu, SWT.NONE);
            setupWithMenuItem.setText(dbversionsList.get(i).getTitle());
            setupWithMenuItem.addSelectionListener(new SelectionAdapter() {
                public void widgetSelected(final SelectionEvent event) {
                    int index = profile_table.getSelectionIndex();
                    if (index != -1) {
                        Profile prof = profilesList.get(index);
                        DosboxVersion dbversion = dbversionsList.get(setupWithSubMenu.indexOf(setupWithMenuItem));
                        try {
                            FileUtils.doRunProfile(prof, dbversion, true);
                        } catch (IOException exc) {
                            GeneralPurposeDialogs.warningMessage(shell, exc.getMessage());
                        }
                    }
                }
            });
        }
    }
    
    private void updateViewLinkSubmenu(Profile prof) {
        MenuItem parent = viewLinkSubMenu.getParentItem();
        viewLinkSubMenu.dispose();
        viewLinkSubMenu = new Menu(parent);
        parent.setMenu(viewLinkSubMenu);
        
        if (prof != null) {
            for (int i = 0; i < prof.getLinks().length; i++) {
                String link = prof.getLink(i);
                if (link.length() > 0) {
                    final MenuItem linkMenuItem = new MenuItem(viewLinkSubMenu, SWT.NONE);
                    String url = link;
                    String tag = link;
                    if (url.indexOf("://") == -1) {
                        url = "file://" + FileUtils.canonicalToData(url).getPath();
                        tag = FileUtils.makeRelativeToDosroot(FileUtils.canonicalToData(tag)).getPath();
                    }
                    String title = prof.getLinkTitle(i);
                    if (title != null && !"".equals(title))
                        tag = title;
                    linkMenuItem.setData(url);
                    linkMenuItem.setText(StringUtils.shorten(tag, 80));
                    linkMenuItem.addSelectionListener(new SelectionAdapter() {
                        public void widgetSelected(final SelectionEvent event) {
                            PlatformUtils.openForBrowsing((String)linkMenuItem.getData());
                        }
                    });
                }
            }
        }
    }
    
    private void doExportProfilesList() {
        new ExportListDialog(shell, dbversionsList, profilesList).open();
    }
}
